#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Validity():
  def __init__(self, fromLane = None, toLane = None) -> None:
    self.from_lane = fromLane
    self.to_lane = toLane

  def parse_from(self, raw_validity):
    if raw_validity:
      self.from_lane = raw_validity.attrib.get('fromLane')
      self.to_lane = raw_validity.attrib.get('toLane')

class Dependency():
  def __init__(self, id = None, type = None):
    self.id = id
    self.type = type

  def parse_from(self, raw_dependency):
    if raw_dependency:
      self.id = raw_dependency.attrib.get('id')
      self.type = raw_dependency.attrib.get('type')

class Signal():
  def __init__(self, s=None, t=None, s_id=None, name=None, dynamic=None,
      orientation=None, pitch = None, roll = None, zOffset=None, country=None,
      countryRevision=None, s_type=None, subtype=None, text = None, value=None,
      unit=None, height=None, width=None, hOffset=None):
    self.s = s
    self.t = t
    self.id = s_id
    self.name = name
    self.dynamic = dynamic
    self.orientation = orientation
    self.pitch = pitch
    self.roll = roll

    self.zOffset = zOffset
    self.country = country
    self.countryRevision = countryRevision
    self.type = s_type
    self.subtype = subtype
    self.text = text
    self.value = value
    self.unit = unit
    self.height = height
    self.width = width
    self.hOffset = hOffset

    self.validity = Validity()
    self.dependency = Dependency()

  def parse_from(self, raw_signal):
    self.s = raw_signal.attrib.get('s')
    self.t = raw_signal.attrib.get('t')
    self.id = raw_signal.attrib.get('id')
    self.name = raw_signal.attrib.get('name')
    self.dynamic = raw_signal.attrib.get('dynamic')
    self.orientation = raw_signal.attrib.get('orientation') # "+" "-" "none"
    self.pitch = raw_signal.attrib.get('pitch') # rad
    self.roll = raw_signal.attrib.get('roll') # rad
    self.zOffset = raw_signal.attrib.get('zOffset')
    self.country = raw_signal.attrib.get('country')
    self.countryRevision = raw_signal.attrib.get('countryRevision')
    self.type = raw_signal.attrib.get('type')
    self.subtype = raw_signal.attrib.get('subtype')
    self.text = raw_signal.attrib.get('text')
    self.value = raw_signal.attrib.get('value')
    self.unit = raw_signal.attrib.get('unit')
    self.height = raw_signal.attrib.get('height')
    self.width = raw_signal.attrib.get('width')
    self.hOffset = raw_signal.attrib.get('hOffset')

    raw_validity = raw_signal.find("validity")
    self.validity.parse_from(raw_validity)
    raw_dependency = raw_signal.find("dependency")
    self.dependency.parse_from(raw_dependency)

  def is_traffic_light(self):
    # Todo(zero): Need add more country code, ref to "12. Signals"
    if self.country == "OpenDRIVE":
      if self.type == "1000001" and self.subtype == "-1":
        return True
    return False

class SignalReference():
  def __init__(self, id = None, orientation = None, s = None, t = None):
    self.id = id
    self.orientation = orientation
    self.s = s
    self.t = t

    self.validity = Validity()

  def parse_from(self, raw_signal_reference):
    self.id = raw_signal_reference.attrib.get('id')
    self.orientation = raw_signal_reference.attrib.get('orientation')
    self.s = raw_signal_reference.attrib.get('s')
    self.t = raw_signal_reference.attrib.get('t')

    raw_validity = raw_signal_reference.find("validity")
    self.validity.parse_from(raw_validity)


class Signals():
  def __init__(self):
    self.signals = []
    self.signal_references = []

  def parse_from(self, raw_signals):
    if raw_signals is None:
      return

    for raw_signal in raw_signals.iter('signal'):
      signal = Signal()
      signal.parse_from(raw_signal)
      self.signals.append(signal)

    for raw_signal_reference in raw_signals.iter('signalReference'):
      signal_reference = SignalReference()
      signal_reference.parse_from(raw_signal_reference)
      self.signal_references.append(signal_reference)



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###############################################################################
# Copyright 2017 The Apollo Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################
"""Protobuf utils."""

import traceback

import google.protobuf.text_format as text_format


def write_pb_to_text_file(topic_pb, file_path):
    """write pb message to file"""
    file_name = "{}.txt".format(file_path)
    with open(file_name, 'w') as f:
        f.write(str(topic_pb))
        print("File success saved in: {}".format(file_name))

def write_pb_to_bin_file(topic_pb, file_path):
    """write pb message to file

    Args:
        topic_pb ([object]): [protobuf file]
        file_path ([string]): [file path]
    """
    file_name = "{}.bin".format(file_path)
    with open(file_name, 'wb') as f:
        f.write(topic_pb.SerializeToString())
        print("File success saved in: {}".format(file_name))


def get_pb_from_text_file(filename, pb_value):
    """Get a proto from given text file."""
    with open(filename, 'r') as file_in:
        return text_format.Merge(file_in.read(), pb_value)


def get_pb_from_bin_file(filename, pb_value):
    """Get a proto from given binary file."""
    with open(filename, 'rb') as file_in:
        pb_value.ParseFromString(file_in.read())
    return pb_value


def get_pb_from_file(filename, pb_value):
    """Get a proto from given file by trying binary mode and text mode."""
    try:
        return get_pb_from_bin_file(filename, pb_value)
    except:
        try:
            return get_pb_from_text_file(filename, pb_value)
        except:
            traceback.print_exc()
            print('Error: Cannot parse %s as binary or text proto', filename)
    return None


def flatten(pb_value, selectors):
    """
    Get a flattened tuple from pb_value. Selectors is a list of sub-fields.

    Usage:
    For a pb_value of:
        total_pb = {
            me: { name: 'myself' }
            children: [{ name: 'child0' }, { name: 'child1' }]
        }
    my_name, child0_name = flatten(total_pb, ['me.name', 'children[0].name'])
    # You get (my_name='myself', child0_name='child0')

    children_names = flatten(total_pb, 'children.name')
    # You get (children_names=['child0', 'child1'])
    """

    def __select_field(val, field):
        if hasattr(val, '__len__'):
            # Flatten repeated field.
            return [__select_field(elem, field) for elem in val]
        if not field.endswith(']'):
            # Simple field.
            return val.__getattribute__(field)
        # field contains index: "field[index]".
        field, index = field.split('[')
        val = val.__getattribute__(field)
        index = int(index[:-1])
        return val[index] if index < len(val) else None

    def __select(val, selector):
        for field in selector.split('.'):
            val = __select_field(val, field)
            if val is None:
                return None
        return val

    # Return the single result for single selector.
    if isinstance(selectors, str):
        return __select(pb_value, selectors)
    # Return tuple result for multiple selectors.
    return tuple((__select(pb_value, selector) for selector in selectors))



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Elevation:
  def __init__(self, s = None, a = None, b = None, c = None, d = None):
    self.s = s
    self.a = a
    self.b = b
    self.c = c
    self.d = d

  def parse_from(self, raw_elevation):
    self.s = float(raw_elevation.attrib.get('s'))
    self.a = float(raw_elevation.attrib.get('a'))
    self.b = float(raw_elevation.attrib.get('b'))
    self.c = float(raw_elevation.attrib.get('c'))
    self.d = float(raw_elevation.attrib.get('d'))


class ElevationProfile:
  def __init__(self):
    self.elevations = []

  def add_elevation(self, elevation):
    self.elevations.append(elevation)

  def parse_from(self, raw_elevation_profile):
    if not raw_elevation_profile:
      return

    for raw_elevation in raw_elevation_profile.iter('elevation'):
      elevation = Elevation()
      elevation.parse_from(raw_elevation)
      self.add_elevation(elevation)

  def get_elevation_by_s(self, s):
    i = len(self.elevations) - 1
    while i >= 0:
      elevation = self.elevations[i]
      if s >= elevation.s:
        a, b, c, d = elevation.a, elevation.b, elevation.c, elevation.d
        ds = s - elevation.s
        elev = a + b*ds + c*ds*ds + d*ds**3
        return elev
      i -= 1
    return 0.0


class LateralProfile:
  def __init__(self):
    self.superelevations = []

  def add_elevation(self, elevation):
    self.superelevations.append(elevation)

  def parse_from(self, raw_lateral_profile):
    if not raw_lateral_profile:
      return

    for raw_elevation in raw_lateral_profile.iter('superelevation'):
      elevation = Elevation()
      elevation.parse_from(raw_elevation)
      self.add_elevation(elevation)

  def get_superelevation_by_s(self, s):
    i = len(self.superelevations) - 1
    while i >= 0:
      superelevation = self.superelevations[i]
      if s >= superelevation.s:
        a = superelevation.a
        b = superelevation.b
        c = superelevation.c
        d = superelevation.d
        ds = s - superelevation.s
        elev = a + b*ds + c*ds*ds + d*ds**3
        return elev
      i -= 1
    return 0.0



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imap.global_var as global_var

from imap.lib.opendrive.common import convert_speed
from imap.lib.opendrive.plan_view import PlanView
from imap.lib.opendrive.profile import ElevationProfile, LateralProfile
from imap.lib.opendrive.lanes import Lanes
from imap.lib.opendrive.signals import Signals


GEOMETRY_SKIP_LENGTH = 0.01


# Type
class Speed:
  def __init__(self, max_speed = None, unit = None):
    self.max_speed = max_speed
    self.unit = unit

  def parse_from(self, raw_speed):
    if raw_speed is None:
      return
    raw_max_speed = raw_speed.attrib.get('max')
    self.unit = raw_speed.attrib.get('unit')
    if raw_max_speed == 'no limit' or raw_max_speed == 'undefined':
      self.max_speed = 0
    else:
      self.max_speed = convert_speed(raw_max_speed, self.unit)


class RoadType:
  def __init__(self, s = None, road_type = None):
    self.s = s
    self.road_type = road_type
    self.speed = Speed()

  def parse_from(self, raw_road_type):
    if raw_road_type is not None:
      self.s = raw_road_type.attrib.get('s')
      self.road_type = raw_road_type.attrib.get('type')

      raw_speed = raw_road_type.find('speed')
      self.speed.parse_from(raw_speed)

# Link
class RoadLink:
  def __init__(self, element_type = None, element_id = None, \
               contact_point = None):
    self.element_type = element_type
    self.element_id = element_id
    self.contact_point = contact_point

  def parse_from(self, raw_data):
    if raw_data is not None:
      self.element_type = raw_data.attrib.get('elementType')
      self.element_id = raw_data.attrib.get('elementId')
      self.contact_point = raw_data.attrib.get('contactPoint')

class Link:
  def __init__(self, predecessor = None, successor = None):
    self.predecessor = RoadLink()
    self.successor = RoadLink()

    # private
    self.predecessor_road = None
    self.successor_road = None
    self.predecessor_junction = None
    self.successor_junction = None

  def parse_from(self, raw_link):
    if raw_link is not None:
      raw_predecessor = raw_link.find('predecessor')
      self.predecessor.parse_from(raw_predecessor)

      raw_successor = raw_link.find('successor')
      self.successor.parse_from(raw_successor)

# Road
class Road:
  def __init__(self, name = None, length = None, road_id = None, \
               junction_id = None):
    self.name = name
    self.length = length
    self.road_id = road_id
    self.junction_id = junction_id

    self.link = Link()
    self.road_type = RoadType()
    self.plan_view = PlanView()
    self.elevation_profile = ElevationProfile()
    self.lateral_profile = LateralProfile()
    self.lanes = Lanes()
    self.signals = Signals()

    # private
    self.reference_line = []


  def generate_lane_boundary(self):
    for lane_section in self.lanes.lane_sections:
      lane_section.left
      lane_section.right


  def post_processing(self):
    # add length
    for idx in range(len(self.lanes.lane_sections) - 1):
      self.lanes.lane_sections[idx].end_s = self.lanes.lane_sections[idx+1].s
    self.lanes.lane_sections[-1].end_s = self.length

    # add neighbor
    for lane_section in self.lanes.lane_sections:
      lane_section.add_neighbors()


  def parse_from(self, raw_road):
    self.name = raw_road.attrib.get('name')
    self.length = float(raw_road.attrib.get('length'))
    self.road_id = raw_road.attrib.get('id')
    self.junction_id = raw_road.attrib.get('junction')

    raw_link = raw_road.find('link')
    self.link.parse_from(raw_link)

    raw_road_type = raw_road.find('type')
    self.road_type.parse_from(raw_road_type)

    # reference line
    raw_plan_view = raw_road.find('planView')
    assert raw_plan_view is not None, \
        "Road {} has no reference line!".format(self.road_id)
    self.plan_view.parse_from(raw_plan_view)

    # elevationProfile
    raw_elevation_profile = raw_road.find('elevationProfile')
    self.elevation_profile.parse_from(raw_elevation_profile)

    # lateralProfile
    raw_lateral_profile = raw_road.find('lateralProfile')
    self.lateral_profile.parse_from(raw_lateral_profile)

    # lanes
    raw_lanes = raw_road.find('lanes')
    assert raw_lanes is not None, "Road {} has no lanes!".format(self.road_id)
    self.lanes.parse_from(raw_lanes)

    # signals
    raw_signals = raw_road.find('signals')
    self.signals.parse_from(raw_signals)

    # post processing
    self.post_processing()

  def generate_reference_line(self):
    for geometry in self.plan_view.geometrys:
      if geometry.length < GEOMETRY_SKIP_LENGTH:
        continue

      sampling_length = global_var.get_element_value("sampling_length")
      points = geometry.sampling(sampling_length, self.elevation_profile)
      self.reference_line.extend(points)

    assert len(self.reference_line) != 0, \
        "Road {} reference line is empty!".format(self.road_id)

  def add_offset_to_reference_line(self):
    for idx in range(len(self.reference_line)):
      if self.lanes.have_offset():
        offset = self.lanes.get_offset_by_s(self.reference_line[idx].s)
        self.reference_line[idx].shift_t(offset)

  def add_origin_to_reference_line(self, origin_x, origin_y):
    for idx in range(len(self.reference_line)):
      self.reference_line[idx].x += origin_x
      self.reference_line[idx].y += origin_y

  def process_lanes(self):
    # generate boundary
    self.lanes.process_lane_sections(self.reference_line)


  def get_cross_section(self, relation):
    return self.lanes.get_cross_section(relation)
  
  def get_cross_section_boundry(self, relation):
    #TODO:: junction boundry
    # <junction.connectingRoad.id -> road.<planView>.geometry.提取
    return self.lanes.get_cross_section_points(relation)



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import imap.global_var as global_var
import modules.map.proto as proto

import matplotlib.pyplot as plt

# Elements with changed attributes
changed_artist = []

def clear_properties():
  for artist in changed_artist:
    artist.set_color('g')
    artist.set_label('')

def set_properties(obj, color, label):
  obj.set_color(color)
  obj.set_label(label)

def show_lane_detail(line, lane):
  set_properties(line, 'red', "cur_lane: " + lane.id.id)
  changed_artist.append(line)
  # pre lanes
  for predecessor_id in lane.predecessor_id:
    line = global_var.get_element_value(predecessor_id.id)
    if line:
      set_properties(line, 'cyan', "pre_lane: " + predecessor_id.id)
      changed_artist.append(line)
  # successor lanes
  for successor_id in lane.successor_id:
    line = global_var.get_element_value(successor_id.id)
    if line:
      set_properties(line, 'purple', "suc_lane: " + successor_id.id)
      changed_artist.append(line)

  # lane.left_neighbor_forward_lane_id
  # lane.right_neighbor_forward_lane_id
  # lane.left_neighbor_reverse_lane_id
  # lane.right_neighbor_reverse_lane_id

def on_click(event):
  pass

def on_pick(event):
  # 1. clear preview label first
  clear_properties()

  # 2. find event.artist
  obj = global_var.get_artist_value(event.artist)
  # print(event.artist)
  if isinstance(obj, proto.map_lane_pb2.Lane):
    show_lane_detail(event.artist, obj)

  # 3. redraw
  plt.legend()
  plt.draw()

def on_press(event):
  pass

def on_release(event):
  pass

def on_motion(event):
  pass



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import xml.etree.ElementTree as ET

from imap.lib.opendrive.header import Header
from imap.lib.opendrive.road import Road
from imap.lib.opendrive.junction import Junction


class Map:
  def __init__(self):
    self.header = Header()
    self.roads = {}
    self.junctions = {}

  def post_process(self):
    # add link
    for road_id, road in self.roads.items():
      # Todo(zero):
      if road.link.predecessor.element_type == "junction":
        road.link.predecessor_junction = \
            self.junctions[road.link.predecessor.element_id]
      elif road.link.predecessor.element_type == "road":
        road.link.predecessor_road = self.roads[road.link.predecessor.element_id]

      if road.link.successor.element_type == "junction":
        road.link.successor_junction = \
            self.junctions[road.link.successor.element_id]
      elif road.link.successor.element_type == "road":
        road.link.successor_road = self.roads[road.link.successor.element_id]

      # add connect relation in junctions
      if road.junction_id != "-1":
        successor_id = road.link.successor.element_id
        if not self.junctions[road.junction_id].is_incoming_road( \
                  successor_id, road.road_id):
          if successor_id not in self.junctions[road.junction_id].predecessor_dict:
            self.junctions[road.junction_id].predecessor_dict[successor_id] = []
          self.junctions[road.junction_id].predecessor_dict[successor_id]. \
              append([road, "successor"])

        predecessor_id = road.link.predecessor.element_id
        if not self.junctions[road.junction_id].is_incoming_road( \
                  predecessor_id, road.road_id):
          if predecessor_id not in self.junctions[road.junction_id].predecessor_dict:
            self.junctions[road.junction_id].predecessor_dict[predecessor_id] = []
          self.junctions[road.junction_id].predecessor_dict[predecessor_id]. \
              append([road, "predecessor"])

      # add connected roads to junctions
      if road.link.predecessor.element_type == "junction":
        junction_id = road.link.predecessor.element_id
        connected_roads = self.junctions[junction_id].connected_roads
        connected_roads.append([road, "predecessor"])
      if road.link.successor.element_type == "junction":
        junction_id = road.link.successor.element_id
        connected_roads = self.junctions[junction_id].connected_roads
        connected_roads.append([road, "successor"])

    # add junction link
    for junction_id, junction in self.junctions.items():
      for connection in junction.connections:
        connection.incoming_road_obj = self.roads[connection.incoming_road]
        connection.connecting_road_obj = self.roads[connection.connecting_road]

  def parse_roads(self, raw_roads):
    for raw_road in raw_roads:
      road = Road()
      road.parse_from(raw_road)
      self.roads[road.road_id] = road


  def parse_junctions(self, raw_junctions):
    if not raw_junctions:
      return

    for raw_junction in raw_junctions:
      junction = Junction()
      junction.parse_from(raw_junction)
      self.junctions[junction.junction_id] = junction

  def load(self, filename):
    tree = ET.parse(filename)
    root = tree.getroot()
    assert root is not None, "Map XML failed!"
    assert root.tag == 'OpenDRIVE'

    # 1. header
    raw_header = root.find('header')
    assert raw_header is not None, "Open drive map missing header"
    self.header.parse_from(raw_header)

    # 3. junction.
    raw_junctions = root.findall('junction')
    self.parse_junctions(raw_junctions)

    # 2. road
    raw_roads = root.findall('road')
    assert raw_roads is not None, "Open drive map missing roads"
    self.parse_roads(raw_roads)

    self.post_process()



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
import copy


class Vector3d:
  def __init__(self, x, y, z):
    self.x = x
    self.y = y
    self.z = z

  def dot_product(self, other) -> float:
    return self.x*other.x + self.y*other.y + self.z*other.z

  def cross_product(self, other):
    x = self.y*other.z - self.z*other.y
    y = self.z*other.x - self.x*other.z
    z = self.x*other.y - self.y*other.x
    return Vector3d(x, y, z)

  def normalize(self):
    length = self.length()
    if length != 0:
      self.x /= length
      self.y /= length
      self.z /= length
    return self

  def length(self) -> float:
    return math.sqrt(self.x**2 + self.y**2 + self.z**2)

  def __add__(self, other):
    self.x += other.x
    self.y += other.y
    self.z += other.z
    return self

  def __sub__(self, other):
    self.x -= other.x
    self.y -= other.y
    self.z -= other.z
    return self

  def __mul__(self, ratio):
    self.x *= ratio
    self.y *= ratio
    self.z *= ratio
    return self

  def __truediv__(self, ratio):
    self.x /= ratio
    self.y /= ratio
    self.z /= ratio
    return self

  def __neg__(self):
    self.x = -self.x
    self.y = -self.y
    self.z = -self.z
    return self

  def __str__(self):
    return "Vector3d x: {}, y: {}, z: {}".format(self.x, self.y, self.z)

class Point3d:
  def __init__(self, x, y, z, s):
    self.x = x
    self.y = y
    self.z = z
    self.s = s

  def set_rotate(self, yaw = 0.0, roll = 0.0, pitch = 0.0):
    self.yaw = yaw
    self.roll = roll
    self.pitch = pitch

  def shift_t(self, offset):
    vec_x = Vector3d(math.cos(self.yaw), math.sin(self.yaw), 0)
    vec_z = Vector3d(0, math.sin(self.roll), math.cos(self.roll))
    normal_xz = vec_x.cross_product(vec_z)
    vec_y = -normal_xz.normalize() * offset

    self.x += vec_y.x
    self.y += vec_y.y
    self.z += vec_y.z

  def __str__(self):
    return "Point3d x: {}, y: {}, z: {}, s: {}, heading: {}".format(self.x, \
        self.y, self.z, self.s, self.yaw)


def shift_t(point3d, offset):
  npoint = copy.deepcopy(point3d)
  npoint.shift_t(offset)

  return npoint

def calc_length(points):
  length = 0
  if math.fabs(points[0].yaw - points[-1].yaw) < 0.01:
    # straight line
    p, q = points[0], points[-1]
    x = q.x - p.x
    y = q.y - p.y
    z = q.z - p.z
    length = math.sqrt(x**2 + y**2 + z**2)
  else:
    for p, q in zip(points, points[1:]):
      x = q.x - p.x
      y = q.y - p.y
      z = q.z - p.z
      length += math.sqrt(x**2 + y**2 + z**2)
  return length


if __name__ == '__main__':
  vec_x = Vector3d(0.9201668879354276, -0.3915263699257437, 0)
  vec_z = Vector3d(0,0,1)
  normal_xz = vec_x.cross_product(vec_z)
  print(normal_xz.normalize())



#!/usr/bin/env python

# Copyright 2024 porter.pan <porter.pan@outlook.com>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imap.global_var as global_var
import imap.lib.proto_utils as proto_utils
from modules.map.proto import map_pb2

from matplotlib.patches import Polygon

class Map:
  def __init__(self):
    self.map_pb = map_pb2.Map()

  def load(self, map_file_name):
    res = proto_utils.get_pb_from_file(map_file_name, self.map_pb)
    return res is not None

  def save(self, map_output_file):
    proto_utils.write_pb_to_text_file(self.map_pb, map_output_file)

  def draw_roads(self, ax, road_ids):
    pass

  def draw_lanes(self, ax, lane_id):
    for lane in self.map_pb.lane:
      if lane.id.id == lane_id:
        self._draw_lane_central(lane, ax, 'r', 1)
      else:
        self._draw_lane_central(lane, ax, 'g', 0.5)

  def draw_junctions(self, ax, junction_ids):
    for junction in self.map_pb.junction:
      if len(junction_ids) == 0 or junction.id.id in junction_ids:
        self._draw_polygon_boundary(junction.polygon, ax, 'c')
        # self._draw_polygon(junction.polygon, ax, 'c')

  def draw_signals(self, ax):
    for signal in self.map_pb.signal:
      for stop_line in signal.stop_line:
        for curve in stop_line.segment:
          self._draw_stop_line(curve.line_segment, ax, "tomato")

  def draw_crosswalks(self, ax):
    for crosswalk in self.map_pb.crosswalk:
      # self._draw_polygon_boundary(crosswalk.polygon, ax, "red")
      self._draw_polygon(crosswalk.polygon, ax, 'c')

  def draw_stop_signs(self, ax):
    for stop_sign in self.map_pb.stop_sign:
      for stop_line in stop_sign.stop_line:
        for curve in stop_line.segment:
          self._draw_stop_line(curve.line_segment, ax, "tomato")

  def draw_yields(self, ax):
    pass
    # for yield_sign in self.map_pb.yield:
    #   for stop_line in yield_sign.stop_line:
    #     for curve in stop_line.segment:
    #       self._draw_stop_line(curve.line_segment, ax, "yellow")

  def draw_clear_areas(self, ax):
    pass

  def draw_overlaps(self, ax):
    pass

  def draw_speed_bumps(self, ax):
    pass

  def draw_parking_spaces(self, ax):
    pass

  def draw_pnc_junctions(self, ax):
    pass


  @staticmethod
  def _draw_lane_boundary(lane, ax, color_val):
    """draw boundary"""
    for curve in lane.left_boundary.curve.segment:
      if curve.HasField('line_segment'):
        px = []
        py = []
        for p in curve.line_segment.point:
            px.append(float(p.x))
            py.append(float(p.y))
        ax.plot(px, py, ls='-', c=color_val, alpha=0.5, picker=True)
    for curve in lane.right_boundary.curve.segment:
      if curve.HasField('line_segment'):
          px = []
          py = []
          for p in curve.line_segment.point:
              px.append(float(p.x))
              py.append(float(p.y))
          ax.plot(px, py, ls='-', c=color_val, alpha=0.5, picker=True)

  @staticmethod
  def _draw_lane_central(lane, ax, color_val, alpha_val = 0.5):
    for curve in lane.central_curve.segment:
      if curve.HasField('line_segment'):
        px = []
        py = []
        for p in curve.line_segment.point:
          px.append(float(p.x))
          py.append(float(p.y))
        line2d, = ax.plot(px, py, ls='-', linewidth=5, c=color_val, alpha=alpha_val, picker=True)

        # add data to global_var
        global_var.set_artist_value(line2d, lane)
        global_var.set_element_vaule(lane.id.id, line2d)

  @staticmethod
  def _draw_polygon_boundary(polygon, ax, color_val):
    px = []
    py = []
    for point in polygon.point:
      px.append(point.x)
      py.append(point.y)

    if px:
      px.append(px[0])
      py.append(py[0])

    ax.plot(px, py, ls='-', linewidth=2, c=color_val, alpha=0.5, picker=True)

  @staticmethod
  def _draw_polygon(polygon, ax, color_val):
    # todo(zero): need to fix
    pxy = []
    for point in polygon.point:
      pxy.append([point.x, point.y])
    polygon = Polygon(pxy, True)
    ax.add_patch(polygon)

  @staticmethod
  def _draw_stop_line(line_segment, ax, color_val):
    px = []
    py = []
    for p in line_segment.point:
      px.append(float(p.x))
      py.append(float(p.y))
    ax.plot(px, py, 'o-', linewidth=1, c=color_val, picker=True)


if __name__ == '__main__':
  pass



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# hull.py
# Graham Scan - Tom Switzer <thomas.switzer@gmail.com>

import math
from functools import reduce

TURN_LEFT, TURN_RIGHT, TURN_NONE = (1, -1, 0)


def cmp(a, b):
    return (a > b) - (a < b)

def turn(p, q, r):
    return cmp((q[0] - p[0])*(r[1] - p[1]) - (r[0] - p[0])*(q[1] - p[1]), 0)

def _keep_left(hull, r):
    while len(hull) > 1 and turn(hull[-2], hull[-1], r) != TURN_LEFT:
            hull.pop()
    if not len(hull) or hull[-1] != r:
        hull.append(r)
    return hull

def convex_hull(points):
    """Returns points on convex hull of an array of points in CCW order."""
    points = sorted(points)
    l = reduce(_keep_left, points, [])
    u = reduce(_keep_left, reversed(points), [])
    return l.extend(u[i] for i in range(1, len(u) - 1)) or l

def aabb_box(points):
    if not points:
        return []
    x_min, y_min = points[0]
    x_max, y_max = points[0]
    for x, y in points:
        x_min = min(x_min, x)
        y_min = min(y_min, y)
        x_max = max(x_max, x)
        y_max = max(y_max, y)
    if math.isclose(x_min, x_max) or math.isclose(y_min, y_max):
        return []
    return [[x_min, y_min], [x_min, y_max], [x_max, y_max], [x_max, y_min]]

if __name__ == '__main__':
    points = [[0,0],[0,1],[1,0],[1,1],[2,2]]
    polygon = convex_hull(points)
    for x, y in polygon:
        print("{}, {}".format(x, y))



# -*- coding: utf-8 -*-
#!/usr/bin/env python

# Copyright 2024 porter.pan <porter.pan@outlook.com>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import logging
import sys
from pathlib import Path

import imap.global_var as global_var

from imap.lib.draw import add_editor, show_map
from imap.lib.convertor import Opendrive2Apollo


def convert_map_format(input_path, output_path, relative, junctionbox=False):
    opendrive2apollo = Opendrive2Apollo(input_path, output_path)
    # todo(zero): only lane type is driving add relationship!!!
    opendrive2apollo.set_parameters(only_driving=False)
    opendrive2apollo.convert(relative, junctionbox)
    opendrive2apollo.save_map()


def show_open_drive_map(map_file):
    opendrive2apollo = Opendrive2Apollo(map_file)
    opendrive2apollo.set_parameters(only_driving=False)
    opendrive2apollo.convert()


def main(args=sys.argv):
    parser = argparse.ArgumentParser(
        description="Imap is a tool to display hdmap info on a map.",
        prog="imap_box_up",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog='eg:\n \
    1.show appollo map by id: imap_box_up -m data/borregas_ave.txt -l lane_35 \n \
    2.Format conversion with UTM map(opendrive->appollo): imap_box_up -f -i data/town.xodr -o data/apollo_map.txt \n \
    3.Format conversion with relative map(opendrive->appollo.relative map, recommend): imap_box_up -f -r -i data/town.xodr -o data/apollo_map.txt \n \
    4.if you want convert map junction with box(not recommend): imap_box_up -f -r -b -i data/town.xodr -o data/apollo_map.txt \n \
    Notice:\n \
    1.After running the command imap -m data/your_map_file, nothing display and no errors!!! \n \
        step1:Check the permissions: sudo chmod 777 data/your_map_file \n \
    if step1 no solve the problem, you can try sudo pip install xxx again. \n \
    2.Other Problem you can contact porter.pan@outlook.com \n \
    3.code modified from: https://github.com/daohu527/imap '
        )

    parser.add_argument(
        "-m", "--map", action="store", type=str, required=False,
        help="Specify the map(appollo hdmap) file in txt or binary format, to show")
    parser.add_argument(
        "--save_fig", action="store", type=bool, required=False,
        default=False, help="Whether to save the visualization figure (only for .xodr suffix)")
    parser.add_argument(
        "-z", "--enable_z_axis", action="store", type=bool, required=False,
        default=False, help="Whether to extract z-axis coordination information to apollo hd-map")
    parser.add_argument(
        "-l", "--lane", action="store", type=str, required=False,
        help="Find lane by lane id")

    parser.add_argument(
        "-f", "--format", action="store", type=str, required=False,
        nargs='?', const="0", help="Convert opendrive hdmap to appollo format")
    parser.add_argument(
        "-i", "--input", action="store", type=str, required=False,
        help="opendrive hdmap input path")
    parser.add_argument(
        "-o", "--output", action="store", type=str, required=False,
        help="for save appollo hdmap file output path")
    parser.add_argument(
        "-s", "--sampling", action="store", type=float, required=False,
        default=1.0, help="sampling length")
    parser.add_argument(
        "-d", "--debug", action="store", type=bool, required=False,
        nargs='?', const=True, default=False, help="debug mode")
    parser.add_argument(
        "-r", "--relative", action="store", type=bool, required=False,
        nargs='?', default=False, help="convert opendrive map to appollo map without GIS projection")
    parser.add_argument(
        "-b", "--box", action="store", type=bool, required=False,
        nargs='?', default=False, help="convert opendrive map to appollo map junction with box(not recommend)")
    args = parser.parse_args(args[1:])

    # 1. Init global var
    global_var._init()
    global_var.set_element_vaule("sampling_length", args.sampling)
    global_var.set_element_vaule("debug_mode", args.debug)
    global_var.set_element_vaule("enable_z_axis", args.enable_z_axis)

    # 2. show map
    if args.map is not None:
        map_file = Path(args.map)
        if not map_file.is_file():
            logging.error("File not exist! '{}'".format(args.map))
            return
        suffix = args.map.split(".")[1]
        if suffix == "bin" or suffix == "txt":
            add_editor()
            show_map(args.map, args.lane)
        elif suffix == "xodr":
            global_var.set_element_vaule("need_save_figure", args.save_fig)
            show_open_drive_map(args.map)
        else:
            pass

    # 3. convert opendrive map to apollo
    if args.format is not None:
        map_file = Path(args.input)
        if not map_file.is_file():
            logging.error("File not exist! '{}'".format(args.input))
            return
        convert_map_format(args.input, args.output, args.relative, args.box)

if __name__ == '__main__':
  main(args=sys.argv)


#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math


class Transform:
  def __init__(self, x, y, z, yaw, roll, pitch) -> None:
    self.origin_x = x
    self.origin_y = y
    self.origin_z = z
    self.yaw = yaw
    self.roll = roll
    self.pitch = pitch

  def set_rotate(self, yaw = 0.0, roll = 0.0, pitch = 0.0) -> None:
    self.yaw = yaw
    self.roll = roll
    self.pitch = pitch

  def set_translate(self, x = 0.0, y = 0.0, z = 0.0) -> None:
    self.origin_x = x
    self.origin_y = y
    self.origin_z = z

  def transform(self, x, y, z):
    self.x, self.y, self.z = x, y, z
    self.r_yaw()
    self.r_roll()
    self.r_pitch()
    return self.translate()

  def r_yaw(self):
    x, y = self.x, self.y
    self.x = x * math.cos(self.yaw) - y * math.sin(self.yaw)
    self.y = x * math.sin(self.yaw) + y * math.cos(self.yaw)

  def r_roll(self):
    y, z = self.y, self.z
    self.y = y * math.cos(self.roll) + z * math.sin(self.roll)
    self.z = -y * math.sin(self.roll) + z * math.cos(self.roll)

  def r_pitch(self):
    x, z = self.x, self.z
    self.x = x * math.cos(self.pitch) - z * math.sin(self.pitch)
    self.z = x * math.sin(self.pitch) + z * math.cos(self.pitch)

  def translate(self):
    self.x += self.origin_x
    self.y += self.origin_y
    self.z += self.origin_z
    return self.x, self.y, self.z



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Link:
  def __init__(self, from_id = None, to_id = None):
    self.from_id = from_id
    self.to_id = to_id

  def parse_from(self, raw_lane_link):
    if raw_lane_link is not None:
      self.from_id = raw_lane_link.attrib.get('from')
      self.to_id = raw_lane_link.attrib.get('to')


class Connection:
  def __init__(self, connection_id = None, connection_type = None, \
               incoming_road = None, connecting_road = None, \
               contact_point = None):
    self.connection_id = connection_id
    self.connection_type = connection_type
    self.incoming_road = incoming_road
    self.connecting_road = connecting_road
    self.contact_point = contact_point
    self.lane_links = []

    # private
    self.incoming_road_obj = None
    self.connecting_road_obj = None

  def parse_from(self, raw_connection):
    if raw_connection is None:
      return

    self.connection_id = raw_connection.attrib.get('id')
    self.connection_type = raw_connection.attrib.get('type')
    self.incoming_road = raw_connection.attrib.get('incomingRoad')
    self.connecting_road = raw_connection.attrib.get('connectingRoad')
    self.contact_point = raw_connection.attrib.get('contactPoint')

    for raw_lane_link in raw_connection.iter('laneLink'):
      lane_link = Link()
      lane_link.parse_from(raw_lane_link)
      self.lane_links.append(lane_link)

  def incoming_lane_link(self, road_id, lane_id):
    if self.incoming_road != road_id:
      return None

    for lane_link in self.lane_links:
      if lane_link.from_id == lane_id:
        return lane_link

    return None


class Junction:
  def __init__(self, junction_id = None, name = None, junction_type = None):
    self.junction_id = junction_id
    self.name = name
    self.junction_type = junction_type
    self.connections = []

    # private
    self.predecessor_dict = {}
    self.connected_roads = []

  def add_connection(self, connection):
    self.connections.append(connection)

  def parse_from(self, raw_junction):
    self.junction_id = raw_junction.attrib.get('id')
    self.name = raw_junction.attrib.get('name')
    self.junction_type = raw_junction.attrib.get('type')

    for raw_connection in raw_junction.iter('connection'):
      connection = Connection()
      connection.parse_from(raw_connection)
      self.add_connection(connection)

  def get_predecessors(self, road_id):
    return self.predecessor_dict.get(road_id, [])

  def is_incoming_road(self, incoming_road, connecting_road):
    for connection in self.connections:
      if connection.incoming_road == incoming_road and \
         connection.connecting_road == connecting_road:
        return True
    return False



#!/usr/bin/env python3

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###############################################################################
# Copyright 2022 The Apollo Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################

import pyproj
import math

def utm2latlon(x, y, zone):
  """utm to latlon"""
  proj = pyproj.Proj(proj='utm', zone=zone, ellps='WGS84')
  lon, lat = proj(x, y, inverse=True)
  return lat, lon


def latlon2utm(lat, lon):
  """latlon to utm"""
  zone = latlon2utmzone(lat, lon)
  projector2 = pyproj.Proj(proj='utm', zone=zone, ellps='WGS84')
  x, y = projector2(lon, lat)
  return x, y, zone


def latlon2utmzone(lat, lon):
  """latlon to utm zone"""
  zone_num = math.floor((lon + 180) / 6) + 1
  if 56.0 <= lat < 64.0 and 3.0 <= lon < 12.0:
    zone_num = 32
  if 72.0 <= lat < 84.0:
    if 0.0 <= lon < 9.0:
      zone_num = 31
    elif 9.0 <= lon < 21.0:
      zone_num = 33
    elif 21.0 <= lon < 33.0:
      zone_num = 35
    elif 33.0 <= lon < 42.0:
      zone_num = 37
  return zone_num



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def convert_speed(speed_limit, speed_unit) -> float:
  if speed_unit == 'm/s':
    return float(speed_limit)
  elif speed_unit == 'km/h':
    return float(speed_limit) * 0.277778
  elif speed_unit == 'mph':
    return float(speed_limit) * 0.44704
  else:
    print("Unsupported speed unit: {}".format(speed_unit))
    return float(speed_limit)



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math

from imap.lib.common import Point3d
from imap.lib.odr_spiral import odr_spiral, odr_arc
from imap.lib.polynoms import cubic_polynoms, parametric_cubic_curve
from imap.lib.transform import Transform

class Geometry:
  def __init__(self, s = None, x = None, y = None, hdg = None, length = None):
    self.s = s
    self.x = x
    self.y = y
    self.hdg = hdg
    self.length = length

  def parse_from(self, raw_geometry):
    self.s = float(raw_geometry.attrib.get('s'))
    self.x = float(raw_geometry.attrib.get('x'))
    self.y = float(raw_geometry.attrib.get('y'))
    self.hdg = float(raw_geometry.attrib.get('hdg'))
    self.length = float(raw_geometry.attrib.get('length'))

  def sampling(self, delta_s, elevation_profile):
    sample_count = math.ceil(self.length / delta_s) + 1

    tf = Transform(self.x, self.y, 0, self.hdg, 0, 0)

    points = []
    for i in range(sample_count):
      s, t, h = min(i * delta_s, self.length), 0, 0
      x, y, z = tf.transform(s, t, h)

      absolute_s = self.s + s

      point3d = Point3d(x, y, z, absolute_s)
      point3d.set_rotate(self.hdg)
      points.append(point3d)
    return points

class Spiral(Geometry):
  def __init__(self, s = None, x = None, y = None, hdg = None, length = None, \
               curv_start = None, curv_end = None):
    super().__init__(s, x, y, hdg, length)
    self.curv_start = curv_start
    self.curv_end = curv_end

  def parse_from(self, raw_geometry):
    super().parse_from(raw_geometry)
    raw_spiral = raw_geometry.find('spiral')
    self.curv_start = float(raw_spiral.attrib.get('curvStart'))
    self.curv_end = float(raw_spiral.attrib.get('curvEnd'))

    self.cdot = (self.curv_end - self.curv_start) / self.length
    self.s0_spiral = self.curv_start / self.cdot
    self.s0, self.t0, self.theta0 = odr_spiral(self.s0_spiral, self.cdot)

  def sampling(self, delta_s, elevation_profile):
    sample_count = math.ceil(self.length / delta_s) + 1

    tf = Transform(self.x, self.y, 0, self.hdg - self.theta0, 0, 0)

    points = []
    for i in range(sample_count):
      local_s = min(i * delta_s, self.length)
      s, t, theta = odr_spiral(local_s + self.s0_spiral, self.cdot)
      x, y, z = tf.transform(s - self.s0, t - self.t0, 0.0)
      z = elevation_profile.get_elevation_by_s(local_s)

      absolute_s = self.s + local_s

      point3d = Point3d(x, y, z, absolute_s)
      point3d.set_rotate(self.hdg + theta - self.theta0)
      points.append(point3d)
    return points


class Arc(Geometry):
  def __init__(self, s = None, x = None, y = None, hdg = None, length = None, \
               curvature = None):
    super().__init__(s, x, y, hdg, length)
    self.curvature = curvature

  def parse_from(self, raw_geometry):
    super().parse_from(raw_geometry)
    raw_arc = raw_geometry.find('arc')
    self.curvature = float(raw_arc.attrib.get('curvature'))

  def sampling(self, delta_s, elevation_profile):
    sample_count = math.ceil(self.length / delta_s) + 1
    tf = Transform(self.x, self.y, 0, self.hdg, 0, 0)

    points = []
    for i in range(sample_count):
      local_s = min(i * delta_s, self.length)
      s, t, theta = odr_arc(local_s, self.curvature)
      x, y, z = tf.transform(s, t, 0.0)
      z = elevation_profile.get_elevation_by_s(local_s)

      # get elevation
      absolute_s = self.s + local_s

      point3d = Point3d(x, y, z, absolute_s)
      point3d.set_rotate(self.hdg + theta)
      points.append(point3d)
    return points


class Poly3(Geometry):
  def __init__(self, s = None, x = None, y = None, hdg = None, length = None, \
               a = None, b = None, c = None, d = None):
    super().__init__(s, x, y, hdg, length)
    self.a = a
    self.b = b
    self.c = c
    self.d = d

  def parse_from(self, raw_geometry):
    super().parse_from(raw_geometry)

    raw_poly3 = raw_geometry.find('poly3')
    self.a = float(raw_poly3.attrib.get('a'))
    self.b = float(raw_poly3.attrib.get('b'))
    self.c = float(raw_poly3.attrib.get('c'))
    self.d = float(raw_poly3.attrib.get('d'))

  def sampling(self, delta_s, elevation_profile):
    sample_count = math.ceil(self.length / delta_s) + 1
    tf = Transform(self.x, self.y, 0, self.hdg, 0, 0)

    points = []
    for i in range(sample_count):
      local_s = min(i * delta_s, self.length)
      s, t, theta = cubic_polynoms(self.a, self.b, self.c, self.d, local_s)
      x, y, z = tf.transform(s, t, 0.0)
      z = elevation_profile.get_elevation_by_s(local_s)

      absolute_s = self.s + local_s

      point3d = Point3d(x, y, z, absolute_s)
      point3d.set_rotate(self.hdg + theta)
      points.append(point3d)
    return points


class ParamPoly3(Geometry):
  def __init__(self, s = None, x = None, y = None, hdg = None, length = None, \
               aU = None, bU = None, cU = None, dU = None, \
               aV = None, bV = None, cV = None, dV = None, pRange = None):
    super().__init__(s, x, y, hdg, length)
    self.aU = aU
    self.bU = bU
    self.cU = cU
    self.dU = dU
    self.aV = aV
    self.bV = bV
    self.cV = cV
    self.dV = dV
    self.pRange = pRange

  def parse_from(self, raw_geometry):
    super().parse_from(raw_geometry)
    raw_param_poly3 = raw_geometry.find('paramPoly3')

    self.aU = float(raw_param_poly3.attrib.get('aU'))
    self.bU = float(raw_param_poly3.attrib.get('bU'))
    self.cU = float(raw_param_poly3.attrib.get('cU'))
    self.dU = float(raw_param_poly3.attrib.get('dU'))
    self.aV = float(raw_param_poly3.attrib.get('aV'))
    self.bV = float(raw_param_poly3.attrib.get('bV'))
    self.cV = float(raw_param_poly3.attrib.get('cV'))
    self.dV = float(raw_param_poly3.attrib.get('dV'))
    self.pRange = raw_param_poly3.attrib.get('pRange')

  def sampling(self, delta_s, elevation_profile):
    sample_count = math.ceil(self.length / delta_s) + 1
    tf = Transform(self.x, self.y, 0, self.hdg, 0, 0)

    points = []
    for i in range(sample_count):
      local_s = min(i * delta_s, self.length)
      if self.pRange == "arcLength":
        s, t, theta = parametric_cubic_curve(
          self.aU, self.bU, self.cU, self.dU,
          self.aV, self.bV, self.cV, self.dV, local_s)
      elif self.pRange == "normalized":
        s, t, theta = parametric_cubic_curve(
          self.aU, self.bU, self.cU, self.dU,
          self.aV, self.bV, self.cV, self.dV, local_s/self.length)
      else:
        print("Unsupported pRange type: {}".format(self.pRange))
        return []
      x, y, z = tf.transform(s, t, 0.0)
      z = elevation_profile.get_elevation_by_s(local_s)

      absolute_s = self.s + local_s

      point3d = Point3d(x, y, z, absolute_s)
      point3d.set_rotate(self.hdg + theta)
      points.append(point3d)
    return points


class PlanView:
  def __init__(self):
    self.geometrys = []

  def add_geometry(self, geometry):
    self.geometrys.append(geometry)

  def parse_from(self, raw_plan_view):
    for raw_geometry in raw_plan_view.iter('geometry'):
      if raw_geometry[0].tag == 'line':
        geometry = Geometry()
      elif raw_geometry[0].tag == 'spiral':
        geometry = Spiral()
      elif raw_geometry[0].tag == 'arc':
        geometry = Arc()
      elif raw_geometry[0].tag == 'poly3':  # deprecated in OpenDrive 1.6.0
        geometry = Poly3()
      elif raw_geometry[0].tag == 'paramPoly3':
        geometry = ParamPoly3()
      else:
        # Todo(zero): raise an exception
        print("geometry type not support")

      geometry.parse_from(raw_geometry)
      self.add_geometry(geometry)



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def _init():
  global _artist_map, _element_map
  _artist_map = {}
  _element_map = {}

def set_artist_value(key, value):
  _artist_map[key] = value

def get_artist_value(key):
  return _artist_map.get(key)

def set_element_vaule(key, value):
  _element_map[key] = value

def get_element_value(key):
  return _element_map.get(key)



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math

import imap.global_var as global_var

from imap.lib.common import shift_t, calc_length
from imap.lib.draw import draw_line

from imap.lib.opendrive.common import convert_speed


def binary_search(arr, val):
  left, right = 0, len(arr) - 1
  while left <= right:
    mid = math.floor((left + right)/2)
    if arr[mid] <= val:
      left = mid + 1
    else:
      right = mid - 1
  return left - 1


def is_adjacent(road_marks) -> bool:
  if not road_marks:
    return True

  road_mark = road_marks[0]
  if road_mark.roadmark_type == "botts dots" or \
     road_mark.roadmark_type == "broken broken" or \
     road_mark.roadmark_type == "broken solid" or \
     road_mark.roadmark_type == "broken" or \
     road_mark.roadmark_type == "none":
    return True
  return False


## LaneOffset
class LaneOffset:
  def __init__(self, s = None, a = None, b = None, c = None, d = None):
    self.s = s
    self.a = a
    self.b = b
    self.c = c
    self.d = d

  def parse_from(self, raw_lane_offset):
    self.s = float(raw_lane_offset.attrib.get('s'))
    self.a = float(raw_lane_offset.attrib.get('a'))
    self.b = float(raw_lane_offset.attrib.get('b'))
    self.c = float(raw_lane_offset.attrib.get('c'))
    self.d = float(raw_lane_offset.attrib.get('d'))

# Lane
class LaneLink:
  def __init__(self, link_id = None):
    self.link_id = link_id

  def parse_from(self, raw_data):
    self.link_id = raw_data.attrib.get("id")


class Link:
  def __init__(self):
    self.predecessor = None
    self.successor = None

  def parse_from(self, raw_link):
    if raw_link is not None:
      raw_predecessor = raw_link.find("predecessor")
      if raw_predecessor is not None:
        self.predecessor = LaneLink()
        self.predecessor.parse_from(raw_predecessor)

      raw_successor = raw_link.find("successor")
      if raw_successor is not None:
        self.successor = LaneLink()
        self.successor.parse_from(raw_successor)


class Width:
  def __init__(self, sOffset = None, a = None, b = None, c = None, d = None):
    self.sOffset = sOffset
    self.a = a
    self.b = b
    self.c = c
    self.d = d

  def parse_from(self, raw_width):
    self.sOffset = float(raw_width.attrib.get("sOffset"))
    self.a = float(raw_width.attrib.get("a"))
    self.b = float(raw_width.attrib.get("b"))
    self.c = float(raw_width.attrib.get("c"))
    self.d = float(raw_width.attrib.get("d"))


class RoadMark:
  def __init__(self, sOffset = None, roadmark_type = None, material = None, \
               color = None, width = None, lane_change = None):
    self.sOffset = sOffset
    self.roadmark_type = roadmark_type
    self.material = material
    self.color = color
    self.width = width
    self.lane_change = lane_change

  def parse_from(self, raw_road_mark):
    self.sOffset = float(raw_road_mark.attrib.get("sOffset"))
    self.color = raw_road_mark.attrib.get("color")
    self.roadmark_type = raw_road_mark.attrib.get("type")
    # self.width = float(raw_road_mark.attrib.get("width"))
    self.lane_change = raw_road_mark.attrib.get("lane_change")


class Speed:
  def __init__(self, sOffset = None, raw_max_v = None, unit = None):
    self.sOffset = sOffset
    self.raw_max_v = raw_max_v
    self.unit = unit
    self.max_v = None

  def parse_from(self, raw_speed):
    if raw_speed is not None:
      self.sOffset = float(raw_speed.attrib.get("sOffset"))
      self.raw_max_v = float(raw_speed.attrib.get("max"))
      self.unit = raw_speed.attrib.get("unit")
      self.max_v = convert_speed(self.raw_max_v, self.unit)

class LaneBoundaryType:
  def __init__(self) -> None:
    self.boundary_type = None
    self.color = None

class Lane:
  def __init__(self, lane_id = None, lane_type = None, level = None, \
              direction = None):
    self.lane_id = lane_id
    self.lane_type = lane_type
    self.level = level
    self.link = Link()
    self.widths = []
    self.speed = Speed()
    self.road_marks = []
    self.user_data = None

    # private
    self.direction = direction
    self.length = None
    self.left_neighbor_forward = []
    self.right_neighbor_forward = []
    self.left_neighbor_reverse = []
    self.right_neighbor_reverse = []
    self.left_boundary = []
    self.left_boundary_type = LaneBoundaryType()
    self.right_boundary = []
    self.right_boundary_type = LaneBoundaryType()
    self.center_line = []


  def add_width(self, width):
    self.widths.append(width)

  def add_road_mark(self, road_mark):
    self.road_marks.append(road_mark)

  def parse_from(self, raw_lane):
    self.lane_id = raw_lane.attrib.get('id')
    self.lane_type = raw_lane.attrib.get('type')
    self.level = raw_lane.attrib.get('level')

    # link
    raw_link = raw_lane.find("link")
    self.link.parse_from(raw_link)

    # width
    for raw_width in raw_lane.iter("width"):
      width = Width()
      width.parse_from(raw_width)
      self.add_width(width)

    # speed
    raw_speed = raw_lane.find("speed")
    self.speed.parse_from(raw_speed)

    # roadMark
    for raw_road_mark in raw_lane.iter("roadMark"):
      road_mark = RoadMark()
      road_mark.parse_from(raw_road_mark)
      self.add_road_mark(road_mark)

  def get_width_by_s(self, s):
    idx = binary_search([width.sOffset for width in self.widths], s)
    a = self.widths[idx].a
    b = self.widths[idx].b
    c = self.widths[idx].c
    d = self.widths[idx].d

    ds = s - self.widths[idx].sOffset
    return a + b*ds + c*ds**2 + d*ds**3
# TODO::compute boundary points by referce line
  def generate_boundary(self, left_boundary, start_s):
    self.left_boundary = left_boundary
    for point3d in left_boundary:
      width = self.get_width_by_s(point3d.s - start_s)

      rpoint3d = shift_t(point3d, width * self.direction)
      self.right_boundary.append(rpoint3d)

      cpoint3d = shift_t(point3d, width * self.direction / 2)
      # todo(daohu527): need update point.s?
      self.center_line.append(cpoint3d)

    # cacl lane length
    self.length = calc_length(self.center_line)

    debug_mode = global_var.get_element_value("debug_mode")
    if not debug_mode:
      if self.lane_type == "driving":
        draw_line(self.left_boundary, 'g')
        draw_line(self.right_boundary, 'r')
      else:
        draw_line(self.left_boundary)
        draw_line(self.right_boundary)
    return self.right_boundary

  def generate_boundary_type(self, left_boundary_type) -> str:
    self.left_boundary_type = left_boundary_type
    if self.road_marks:
      self.right_boundary_type.boundary_type = self.road_marks[0].roadmark_type
      self.right_boundary_type.color = self.road_marks[0].color
    return self.right_boundary_type

class LaneSection:
  def __init__(self, s = None, single_side = None):
    self.s = s
    self.single_side = single_side
    self.left = []
    self.center = Lane()
    self.right = []

    # private
    self.end_s = None

  def add_left_lane(self, lane):
    self.left.append(lane)

  def add_right_lane(self, lane):
    self.right.append(lane)

  def parse_from(self, raw_lane_section):
    self.s = float(raw_lane_section.attrib.get('s'))

    self.single_side = bool(raw_lane_section.attrib.get('singleSide'))

    # left
    left = raw_lane_section.find("left")
    if left is not None:
      for raw_lane in left.iter('lane'):
        lane = Lane(direction = 1)
        lane.parse_from(raw_lane)
        self.add_left_lane(lane)

    # center
    center = raw_lane_section.find("center")
    raw_lane = center.find('lane')
    self.center.parse_from(raw_lane)

    # right
    right = raw_lane_section.find("right")
    if right is not None:
      for raw_lane in right.iter('lane'):
        lane = Lane(direction = -1)
        lane.parse_from(raw_lane)
        self.add_right_lane(lane)

  def add_neighbors(self):
    n = len(self.left)
    for idx in range(n):
      if idx+1 < n and self.left[idx+1].lane_type == "driving":
        self.left[idx].left_neighbor_forward.append(self.left[idx+1].lane_id)
      if idx > 0 and self.left[idx-1].lane_type == "driving":
        self.left[idx].right_neighbor_forward.append(self.left[idx-1].lane_id)

    n = len(self.right)
    for idx in range(n):
      if idx+1 < n and self.right[idx+1].lane_type == "driving":
        self.right[idx].right_neighbor_forward.append(self.right[idx+1].lane_id)
      if idx > 0 and self.right[idx-1].lane_type == "driving":
        self.right[idx].left_neighbor_forward.append(self.right[idx-1].lane_id)

    if self.left and self.right and is_adjacent(self.center.road_marks):
      if self.right[0].lane_type == "driving":
        self.left[-1].left_neighbor_reverse.append(self.right[0].lane_id)
      if self.left[-1].lane_type == "driving":
        self.right[0].left_neighbor_reverse.append(self.left[-1].lane_id)

  def process_lane(self, reference_line_in_section):
    left_boundary = reference_line_in_section.copy()
    left_boundary_type = self.center.generate_boundary_type(None)
    # The left lane is opposite to the reference line
    left_boundary.reverse()
    for lane in self.left[::-1]:
      left_boundary = lane.generate_boundary(left_boundary, self.s)
      left_boundary_type = lane.generate_boundary_type(left_boundary_type)

    left_boundary = reference_line_in_section.copy()
    left_boundary_type = self.center.generate_boundary_type(None)
    for lane in self.right:
      left_boundary = lane.generate_boundary(left_boundary, self.s)
      left_boundary_type = lane.generate_boundary_type(left_boundary_type)

  def get_cross_section(self, direction):
    if direction == "start":
      if self.left and self.right:
        leftmost_lane, rightmost_lane = self.left[0], self.right[-1]
        return [leftmost_lane.right_boundary[-1], rightmost_lane.right_boundary[0]]
      elif self.left:
        leftmost_lane, rightmost_lane = self.left[0], self.left[-1]
        return [leftmost_lane.right_boundary[-1], rightmost_lane.left_boundary[-1]]
      elif self.right:
        leftmost_lane, rightmost_lane = self.right[0], self.right[-1]
        return [leftmost_lane.left_boundary[0], rightmost_lane.right_boundary[0]]
      else:
        return []
    elif direction == "end":
      if self.left and self.right:
        leftmost_lane, rightmost_lane = self.left[0], self.right[-1]
        return [leftmost_lane.right_boundary[0], rightmost_lane.right_boundary[-1]]
      elif self.left:
        leftmost_lane, rightmost_lane = self.left[0], self.left[-1]
        return [leftmost_lane.right_boundary[0], rightmost_lane.left_boundary[0]]
      elif self.right:
        leftmost_lane, rightmost_lane = self.right[0], self.right[-1]
        return [leftmost_lane.left_boundary[-1], rightmost_lane.right_boundary[-1]]
      else:
        return []
    else:
      return []
    
  def get_cross_section_points_(self, direction):
    if direction == "start":
      if self.left and self.right:
        leftmost_lane, rightmost_lane = self.left[0], self.right[-1]
        return [leftmost_lane.right_boundary[-1::-1], rightmost_lane.right_boundary[0:]]
      
      elif self.left:
        leftmost_lane, rightmost_lane = self.left[0], self.left[-1]
        return [leftmost_lane.right_boundary[-1::-1], rightmost_lane.left_boundary[-1::-1]]
      
      elif self.right:
        leftmost_lane, rightmost_lane = self.right[0], self.right[-1]
        return [leftmost_lane.left_boundary[0:], rightmost_lane.right_boundary[0:]]
      else:
        return []
    elif direction == "end":
      if self.left and self.right:
        leftmost_lane, rightmost_lane = self.left[0], self.right[-1]
        return [leftmost_lane.right_boundary[0:], rightmost_lane.right_boundary[-1::-1]]
      elif self.left:
        leftmost_lane, rightmost_lane = self.left[0], self.left[-1]
        return [leftmost_lane.right_boundary[0:], rightmost_lane.left_boundary[0:]]
      elif self.right:
        leftmost_lane, rightmost_lane = self.right[0], self.right[-1]
        return [leftmost_lane.left_boundary[-1::-1], rightmost_lane.right_boundary[-1::-1]]
      else:
        return []
    else:
      return []

  def leftmost_boundary(self):
    for lane in self.left:
      if lane.lane_type == "driving":
        return lane.right_boundary[::-1], lane.length

    for lane in self.right:
      if lane.lane_type == "driving":
        return lane.left_boundary, lane.length

    return [], 0

  def rightmost_boundary(self):
    for lane in self.right[::-1]:
      if lane.lane_type == "driving":
        return lane.right_boundary, lane.length

    for lane in self.left[::-1]:
      if lane.lane_type == "driving":
        return lane.left_boundary[::-1], lane.length

    return [], 0

# Lanes
class Lanes:
  def __init__(self):
    self.lane_offsets = []
    self.lane_sections = []

  def add_lane_offset(self, lane_offset):
    self.lane_offsets.append(lane_offset)

  def add_lane_section(self, lane_section):
    self.lane_sections.append(lane_section)

  def parse_from(self, raw_lanes):
    # laneOffset
    for raw_lane_offset in raw_lanes.iter('laneOffset'):
      lane_offset = LaneOffset()
      lane_offset.parse_from(raw_lane_offset)
      self.add_lane_offset(lane_offset)

    # laneSection
    i = 0
    for raw_lane_section in raw_lanes.iter("laneSection"):
      lane_section = LaneSection()
      lane_section.parse_from(raw_lane_section)
      self.add_lane_section(lane_section)
      i += 1

  def have_offset(self):
    return len(self.lane_offsets) != 0

  def get_offset_by_s(self, s):
    idx = binary_search([lane_offset.s for lane_offset in self.lane_offsets], s)
    a = self.lane_offsets[idx].a
    b = self.lane_offsets[idx].b
    c = self.lane_offsets[idx].c
    d = self.lane_offsets[idx].d
    ds = s - self.lane_offsets[idx].s
    return a + b*ds + c*ds**2 + d*ds**3

  def process_lane_sections(self, reference_line):
    for lane_section in self.lane_sections:
      start_s = lane_section.s
      end_s = lane_section.end_s
      reference_line_in_section = []
      for (idx, point3d) in enumerate(reference_line):
        if start_s <= point3d.s <= end_s or \
          (point3d.s < start_s and idx + 1 < len(reference_line) and start_s <= reference_line[idx + 1].s <= end_s) or \
            (end_s < point3d.s and idx - 1 >= 0 and start_s <= reference_line[idx - 1].s <= end_s):
          reference_line_in_section.append(point3d)
      lane_section.process_lane(reference_line_in_section)

  def get_cross_section(self, relation):
    if relation == "predecessor":
      return self.lane_sections[0].get_cross_section("start")
    elif relation == "successor":
      return self.lane_sections[-1].get_cross_section("end")
    else:
      print("Unknown relation!")
      return []

  
      
  def get_cross_section_points(self, relation):
    # qian
    if relation == "predecessor":
      return self.lane_sections[0].get_cross_section("start")
    # hou
    elif relation == "successor":
      return self.lane_sections[-1].get_cross_section("end")
    else:
      print("Unknown relation!")
      return []



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

def cubic_polynoms(a, b, c, d, u):
  v = a + b*u + c*u**2 + d*u**3
  theta = math.atan(b + 2*c*u + 3*d*u**2)
  return u, v, theta

def parametric_cubic_curve(aU, bU, cU, dU, aV, bV, cV, dV, p):
  u = aU + bU*p + cU*p**2 + dU*p**3
  v = aV + bV*p + cV*p**2 + dV*p**3
  theta = math.atan2(bV + 2*cV*p + 3*dV*p**2, bU + 2*cU*p + 3*dU*p**2)
  return u, v, theta



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



class GeoReference:
  def __init__(self):
    self.text = None

  def parse_from(self, raw_geo_reference):
    if raw_geo_reference is not None:
      self.text = raw_geo_reference.text


class Header:
  def __init__(self, rev_major = None, rev_minor = None, name = None,
               version = None, date = None, north = None, south = None, \
               east = None, west = None, vendor = None):
    self.rev_major = rev_major
    self.rev_minor = rev_minor
    self.name = name
    self.version = version
    self.date = date
    self.north = north
    self.south = south
    self.east = east
    self.west = west
    self.vendor = vendor
    self.geo_reference = GeoReference()

  def parse_from(self, raw_header):
    self.rev_major = raw_header.attrib.get('revMajor').encode()
    self.rev_minor = raw_header.attrib.get('revMinor').encode()
    if raw_header.attrib.get('name') is not None:
      self.name = raw_header.attrib.get('name').encode()
    if raw_header.attrib.get('version') is not None:
      self.version = raw_header.attrib.get('version').encode()
    if raw_header.attrib.get('date') is not None:
      self.date = raw_header.attrib.get('date').encode()
    if raw_header.attrib.get('east') is not None:
      self.east = float(raw_header.attrib.get('east'))
    if raw_header.attrib.get('west') is not None:
      self.west = float(raw_header.attrib.get('west'))
    if raw_header.attrib.get('south') is not None:
      self.south = float(raw_header.attrib.get('south'))
    if raw_header.attrib.get('north') is not None:
      self.north = float(raw_header.attrib.get('north'))
    if raw_header.attrib.get('vendor') is not None:
      self.vendor = raw_header.attrib.get('vendor').encode()

    raw_geo_reference = raw_header.find("geoReference")
    self.geo_reference.parse_from(raw_geo_reference)



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import math

from modules.map.proto import map_pb2
from modules.map.proto import map_road_pb2
from modules.map.proto import map_lane_pb2

import imap.global_var as global_var

from imap.lib.opendrive.map import Map
from imap.lib.proto_utils import (
  write_pb_to_text_file,
  write_pb_to_bin_file
)

from imap.lib.draw import draw_line, show
from imap.lib.convex_hull import convex_hull, aabb_box
from imap.lib.proj_helper import latlon2utm
from concave_hull import concave_hull as concave_hull_lib
from scipy.spatial import ConvexHull as ConvexHull_lib
import numpy as np


# Distance between stop line and pedestrian crossing
STOP_LINE_DISTANCE = 1.0


def to_pb_lane_type(open_drive_type):
  if open_drive_type is None:
    return map_lane_pb2.Lane.NONE

  lower_type = open_drive_type.lower()
  if lower_type == 'none':
    return map_lane_pb2.Lane.NONE
  elif lower_type == 'driving':
    return map_lane_pb2.Lane.CITY_DRIVING
  elif lower_type == 'biking':
    return map_lane_pb2.Lane.BIKING
  elif lower_type == 'sidewalk':
    return map_lane_pb2.Lane.SIDEWALK
  elif lower_type == 'parking':
    return map_lane_pb2.Lane.PARKING
  elif lower_type == 'shoulder':
    return map_lane_pb2.Lane.SHOULDER
  elif lower_type == 'border':     # not support
    return map_lane_pb2.Lane.NONE
  elif lower_type == 'stop':       # not support
    return map_lane_pb2.Lane.NONE
  elif lower_type == 'restricted': # not support
    return map_lane_pb2.Lane.NONE
  elif lower_type == 'median':     # not support
    return map_lane_pb2.Lane.NONE
  elif lower_type == 'curb':       # not support
    return map_lane_pb2.Lane.NONE
  elif lower_type == 'exit':       # not support
    return map_lane_pb2.Lane.NONE
  elif lower_type == 'entry':      # not support
    return map_lane_pb2.Lane.NONE
  elif lower_type == 'onramp':     # not support
    return map_lane_pb2.Lane.NONE
  elif lower_type == 'offRamp':    # not support
    return map_lane_pb2.Lane.NONE
  elif lower_type == 'connectingRamp': # not support
    return map_lane_pb2.Lane.NONE
  else:
    logging.info("Unsupported lane type: {}".format(open_drive_type))
    return map_lane_pb2.Lane.NONE

def to_pb_boundary_type(opendrive_boundary_type):
  if (opendrive_boundary_type.boundary_type is None or
      opendrive_boundary_type.color is None):
    return map_lane_pb2.LaneBoundaryType.UNKNOWN

  lower_type = opendrive_boundary_type.boundary_type.lower()
  lower_color = opendrive_boundary_type.color.lower()

  # Table 38. Attributes of the road lanes laneSection lcr lane roadMark element
  # e_roadMarkColor & e_roadMarkType
  if lower_type == 'solid solid' and lower_color == 'yellow':
    return map_lane_pb2.LaneBoundaryType.DOUBLE_YELLOW

  if lower_type == 'broken':
    if lower_color == 'yellow':
      return map_lane_pb2.LaneBoundaryType.DOTTED_YELLOW
    elif lower_color == 'white':
      return map_lane_pb2.LaneBoundaryType.DOTTED_WHITE

  if lower_type == 'solid':
    if lower_color == 'yellow':
      return map_lane_pb2.LaneBoundaryType.SOLID_YELLOW
    elif lower_color == 'white':
      return map_lane_pb2.LaneBoundaryType.SOLID_WHITE

  if lower_type == 'curb':
    return map_lane_pb2.LaneBoundaryType.CURB

  return map_lane_pb2.LaneBoundaryType.UNKNOWN

class Convertor:
  def __init__(self) -> None:
    pass

  def convert(self):
    pass


class Opendrive2Apollo(Convertor):
  def __init__(self, input_file_name, output_file_name = None) -> None:
    self.xodr_map = Map()
    self.xodr_map.load(input_file_name)

    # lhd for saving figure
    self.input_file_name = input_file_name

    self.pb_map = map_pb2.Map()

    self.output_file_name = self._get_file_name(output_file_name)
    # UTM coordinate
    self.origin_x = 0.0
    self.origin_y = 0.0

  def _get_file_name(self, file_name):
    if file_name and file_name.endswith((".txt", ".bin")):
      return file_name.rsplit('.', 1)[0]
    return None

  def set_parameters(self, only_driving = True):
    self.only_driving = only_driving

  def convert_proj_txt(self, proj_txt, relative):
    if proj_txt is None:
      self.pb_map.header.projection.proj = "+proj=utm +zone={} +ellps=WGS84 " \
        "+datum=WGS84 +units=m +no_defs".format(0)
      return

    if '+proj=utm' in proj_txt:
      self.pb_map.header.projection.proj = proj_txt
    else:
      # We want just support +proj=tmerc, but some do not contain this parameter
      for p in proj_txt.split():
        if p.startswith('+lat_0'):
          lat = float(p.split('=')[1])
        elif p.startswith('+lon_0'):
          lon = float(p.split('=')[1])
      if lat is None or lon is None:
        self.pb_map.header.projection.proj = "+proj=utm +zone={} +ellps=WGS84 " \
          "+datum=WGS84 +units=m +no_defs".format(0)
      else:
        self.origin_x, self.origin_y, zone_id = latlon2utm(lat, lon)
        self.pb_map.header.projection.proj = "+proj=utm +zone={} +ellps=WGS84 " \
          "+datum=WGS84 +units=m +no_defs".format(zone_id)
    if relative == True:
      self.origin_x, self.origin_y, zone_id=0,0,0

  def convert_header(self, relative):
    if self.xodr_map.header.version:
      self.pb_map.header.version = self.xodr_map.header.version
    if self.xodr_map.header.date:
      self.pb_map.header.date = self.xodr_map.header.date

    proj_txt = self.xodr_map.header.geo_reference.text
    self.convert_proj_txt(proj_txt, relative)

    # TODO(zero): Inconsistent definitions
    # self.pb_map.header.district = self.xodr_map.header.name
    if self.xodr_map.header.rev_major:
      self.pb_map.header.rev_major = self.xodr_map.header.rev_major
    if self.xodr_map.header.rev_minor:
      self.pb_map.header.rev_minor = self.xodr_map.header.rev_minor
    if self.xodr_map.header.west:
      self.pb_map.header.left = self.xodr_map.header.west
    if self.xodr_map.header.east:
      self.pb_map.header.right = self.xodr_map.header.east
    if self.xodr_map.header.north:
      self.pb_map.header.top = self.xodr_map.header.north
    if self.xodr_map.header.south:
      self.pb_map.header.bottom = self.xodr_map.header.south
    if self.xodr_map.header.vendor:
      self.pb_map.header.vendor = self.xodr_map.header.vendor


  def add_basic_info(self, pb_lane, xodr_road, idx, lane):
    pb_lane.id.id = "road_{}_lane_{}_{}".format(xodr_road.road_id, \
        idx, lane.lane_id)
    pb_lane.type = to_pb_lane_type(lane.lane_type)
    pb_lane.length = lane.length
    # Lane speed first, then road, and finally the default 120km/h
    if lane.speed.max_v:
      pb_lane.speed_limit = lane.speed.max_v
    elif xodr_road.road_type.speed.max_speed:
      pb_lane.speed_limit = xodr_road.road_type.speed.max_speed
    else:
      pb_lane.speed_limit = 33.3
    pb_lane.direction = map_lane_pb2.Lane.FORWARD


  def add_lane_boundary(self, pb_lane, lane):
    # 1. left boundary
    segment = pb_lane.left_boundary.curve.segment.add()
    for point3d in lane.left_boundary:
      point = segment.line_segment.point.add()
      # lhd 2022/12/03 for 3D view
      if global_var.get_element_value("enable_z_axis"):
        point.x, point.y, point.z = point3d.x, point3d.y, point3d.z
      else:
        point.x, point.y = point3d.x, point3d.y
    segment.s = 0
    segment.start_position.x = lane.left_boundary[0].x
    segment.start_position.y = lane.left_boundary[0].y
    segment.start_position.z = lane.left_boundary[0].z
    segment.length = pb_lane.length
    pb_lane.left_boundary.length = pb_lane.length
    pb_boundary_type = to_pb_boundary_type(lane.left_boundary_type)
    boundary_type = pb_lane.left_boundary.boundary_type.add()
    boundary_type.s = 0
    boundary_type.types.append(pb_boundary_type)

    # 2. center line
    segment = pb_lane.central_curve.segment.add()
    for point3d in lane.center_line:
      point = segment.line_segment.point.add()
      # lhd 2022/12/03 for 3D view
      if global_var.get_element_value("enable_z_axis"):
        point.x, point.y, point.z = point3d.x, point3d.y, point3d.z
      else:
        point.x, point.y = point3d.x, point3d.y
    segment.s = 0
    segment.start_position.x = lane.center_line[0].x
    segment.start_position.y = lane.center_line[0].y
    segment.start_position.z = lane.center_line[0].z
    segment.length = pb_lane.length

    # 3. right boundary
    segment = pb_lane.right_boundary.curve.segment.add()
    for point3d in lane.right_boundary:
      point = segment.line_segment.point.add()
      # lhd 2022/12/03 for 3D view
      if global_var.get_element_value("enable_z_axis"):
        point.x, point.y, point.z = point3d.x, point3d.y, point3d.z
      else:
        point.x, point.y = point3d.x, point3d.y
    segment.s = 0
    segment.start_position.x = lane.right_boundary[0].x
    segment.start_position.y = lane.right_boundary[0].y
    segment.start_position.z = lane.right_boundary[0].z
    segment.length = pb_lane.length
    pb_lane.right_boundary.length = pb_lane.length
    pb_boundary_type = to_pb_boundary_type(lane.right_boundary_type)
    boundary_type = pb_lane.right_boundary.boundary_type.add()
    boundary_type.s = 0
    boundary_type.types.append(pb_boundary_type)

  def add_lane_sample(self, pb_lane, lane):
    cur_lane_id = int(lane.lane_id)
    total_s = lane.center_line[0].s
    for point3d in lane.center_line:
      lane_width = lane.get_width_by_s(point3d.s)

      # 1. left sample
      left_sample = pb_lane.left_sample.add()
      left_sample.width = lane_width / 2
      # 2. right sample
      right_sample = pb_lane.right_sample.add()
      right_sample.width = lane_width / 2
      # left lane's should be reverse
      if cur_lane_id > 0:
        left_sample.s = total_s - point3d.s
        right_sample.s = total_s - point3d.s
      else:
        left_sample.s = point3d.s
        right_sample.s = point3d.s


  def add_lane_neighbors(self, pb_lane, xodr_road, idx, lane):
    for lane_id in lane.left_neighbor_forward:
      pb_lane.left_neighbor_forward_lane_id.add().id = \
          "road_{}_lane_{}_{}".format(xodr_road.road_id, idx, lane_id)

    for lane_id in lane.right_neighbor_forward:
      pb_lane.right_neighbor_forward_lane_id.add().id = \
          "road_{}_lane_{}_{}".format(xodr_road.road_id, idx, lane_id)

    for lane_id in lane.left_neighbor_reverse:
      pb_lane.left_neighbor_reverse_lane_id.add().id = \
          "road_{}_lane_{}_{}".format(xodr_road.road_id, idx, lane_id)


  def outcoming_road_relationships(self, pb_lane, lane, predecessors, xodr_road):
    # print("--------")
    for predecessor_road, dirct in predecessors:
      # print("{}->{}".format(predecessor_road.road_id, xodr_road.road_id))
      section_id = len(predecessor_road.lanes.lane_sections) - 1
      # left
      if dirct == "predecessor":
        for predecessor_lane in predecessor_road.lanes.lane_sections[0].left:
          if predecessor_lane.link.predecessor and \
             predecessor_lane.link.predecessor.link_id == lane.lane_id:
            pb_lane.predecessor_id.add().id = "road_{}_lane_{}_{}".format( \
                predecessor_road.road_id, 0, predecessor_lane.lane_id)
      elif dirct == "successor":
        for predecessor_lane in predecessor_road.lanes.lane_sections[section_id].right:
          if predecessor_lane.link.successor and \
            predecessor_lane.link.successor.link_id == lane.lane_id:
            pb_lane.predecessor_id.add().id = "road_{}_lane_{}_{}".format( \
                predecessor_road.road_id, section_id, predecessor_lane.lane_id)
      else:
        print("Unknown direction!")


  def add_junction_relationships(self, pb_lane, xodr_road, lane_section, idx, lane):
    cur_n = len(xodr_road.lanes.lane_sections)
    if idx == 0:
      predecessor_junction = xodr_road.link.predecessor_junction
      if predecessor_junction is not None:
        # incoming_road
        for connection in predecessor_junction.connections:
          lane_link = connection.incoming_lane_link(xodr_road.road_id, \
                          lane.lane_id)
          if lane_link is not None:
            section_id = 0
            pb_lane.successor_id.add().id = "road_{}_lane_{}_{}".format( \
                connection.connecting_road, section_id, lane_link.to_id)
        # outcoming_road
        predecessors = predecessor_junction.get_predecessors(xodr_road.road_id)
        self.outcoming_road_relationships(pb_lane, lane, predecessors, xodr_road)

    if idx == cur_n - 1:
      successor_junction = xodr_road.link.successor_junction
      if successor_junction is not None:
        # incoming_road
        for connection in successor_junction.connections:
          lane_link = connection.incoming_lane_link(xodr_road.road_id, \
                          lane.lane_id)
          if lane_link is not None:
            section_id = 0
            pb_lane.successor_id.add().id = "road_{}_lane_{}_{}".format( \
                connection.connecting_road, section_id, lane_link.to_id)
        # outcoming_road
        predecessors = successor_junction.get_predecessors(xodr_road.road_id)
        self.outcoming_road_relationships(pb_lane, lane, predecessors, xodr_road)


  def add_lane_relationships(self, pb_lane, xodr_road, lane_section, idx, lane):
    cur_n = len(xodr_road.lanes.lane_sections)
    cur_lane_id = int(lane.lane_id)
    # 1. External connection
    if idx == 0:
      # 1.1 predecessor road
      predecessor_road_id = xodr_road.link.predecessor.element_id
      if predecessor_road_id and lane.link.predecessor:
        section_id = 0
        if xodr_road.link.predecessor.contact_point == "start":
          section_id = 0
        elif xodr_road.link.predecessor.contact_point == "end":
          section_id = len(xodr_road.link.predecessor_road.lanes.lane_sections) - 1

        if cur_lane_id < 0:
          pb_lane.predecessor_id.add().id = "road_{}_lane_{}_{}".format( \
              predecessor_road_id, section_id, lane.link.predecessor.link_id)
        elif cur_lane_id > 0:
          pb_lane.successor_id.add().id = "road_{}_lane_{}_{}".format( \
              predecessor_road_id, section_id, lane.link.predecessor.link_id)
    if idx == cur_n - 1:
      # 1.2 successor road
      successor_road_id = xodr_road.link.successor.element_id
      if successor_road_id and lane.link.successor:
        section_id = 0
        if xodr_road.link.successor.contact_point == "start":
          section_id = 0
        elif xodr_road.link.successor.contact_point == "end":
          section_id = len(xodr_road.link.successor_road.lanes.lane_sections) - 1

        if cur_lane_id < 0:
          pb_lane.successor_id.add().id = "road_{}_lane_{}_{}".format( \
              successor_road_id, section_id, lane.link.successor.link_id)
        elif cur_lane_id > 0:
          pb_lane.predecessor_id.add().id = "road_{}_lane_{}_{}".format( \
              successor_road_id, section_id, lane.link.successor.link_id)

    # 2. Internal connection
    if idx > 0 and lane.link.predecessor:
      if cur_lane_id < 0:
        pb_lane.predecessor_id.add().id = "road_{}_lane_{}_{}".format( \
            xodr_road.road_id, idx - 1, lane.link.predecessor.link_id)
      elif cur_lane_id > 0:
        pb_lane.successor_id.add().id = "road_{}_lane_{}_{}".format( \
            xodr_road.road_id, idx - 1, lane.link.predecessor.link_id)
    if idx < cur_n - 1 and lane.link.successor:
      if cur_lane_id < 0:
        pb_lane.successor_id.add().id = "road_{}_lane_{}_{}".format( \
            xodr_road.road_id, idx + 1, lane.link.successor.link_id)
      elif cur_lane_id > 0:
        pb_lane.predecessor_id.add().id = "road_{}_lane_{}_{}".format( \
            xodr_road.road_id, idx + 1, lane.link.successor.link_id)


  def create_lane(self, xodr_road, lane_section, idx, lane):
    if self.only_driving and lane.lane_type != "driving":
      return

    pb_lane = self.pb_map.lane.add()
    self.add_basic_info(pb_lane, xodr_road, idx, lane)
    # add boundary
    self.add_lane_boundary(pb_lane, lane)
    # add lane sample
    self.add_lane_sample(pb_lane, lane)
    # add neighbor
    self.add_lane_neighbors(pb_lane, xodr_road, idx, lane)
    # predecessor road
    self.add_lane_relationships(pb_lane, xodr_road, lane_section, idx, lane)
    self.add_junction_relationships(pb_lane, xodr_road, lane_section, idx, lane)
    return pb_lane


  def add_road_section_curve(self, pb_boundary_edge, boundary, length):
    segment = pb_boundary_edge.curve.segment.add()
    for point3d in boundary:
      point = segment.line_segment.point.add()
      # lhd 2022/12/03 for 3D view
      if global_var.get_element_value("enable_z_axis"):
        point.x, point.y, point.z = point3d.x, point3d.y, point3d.z
      else:
        point.x, point.y = point3d.x, point3d.y
    segment.s = 0
    segment.start_position.x = boundary[0].x
    segment.start_position.y = boundary[0].y
    segment.start_position.z = boundary[0].z
    segment.length = length


  def add_road_section_boundary(self, pb_road_section, lane_section):
    left_boundary_edge = pb_road_section.boundary.outer_polygon.edge.add()
    left_boundary_edge.type = map_road_pb2.BoundaryEdge.LEFT_BOUNDARY
    right_boundary_edge = pb_road_section.boundary.outer_polygon.edge.add()
    right_boundary_edge.type = map_road_pb2.BoundaryEdge.RIGHT_BOUNDARY


    leftmost_boundary, leftmost_length = lane_section.leftmost_boundary()
    rightmost_boundary, rightmost_length = lane_section.rightmost_boundary()

    if not leftmost_boundary or not rightmost_boundary:
      # TODO(zero): No leftmost_boundary and rightmost_boundary?
      return

    self.add_road_section_curve(left_boundary_edge, \
        leftmost_boundary, leftmost_length)
    self.add_road_section_curve(right_boundary_edge, \
        rightmost_boundary, rightmost_length)


  def convert_lane(self, xodr_road, pb_road):
    for idx, lane_section in enumerate(xodr_road.lanes.lane_sections):
      pb_road_section = pb_road.section.add()
      pb_road_section.id.id = str(idx)
      self.add_road_section_boundary(pb_road_section, lane_section)

      for lane in lane_section.left:
        pb_lane = self.create_lane(xodr_road, lane_section, idx, lane)
        # Not driving road is None
        if pb_lane is not None:
          pb_road_section.lane_id.add().id = pb_lane.id.id

      # The last section of road, which used to generate stop lines for
      # traffic lights
      pb_last_section = []
      for lane in lane_section.right:
        pb_lane = self.create_lane(xodr_road, lane_section, idx, lane)
        if pb_lane is not None:
          pb_road_section.lane_id.add().id = pb_lane.id.id
          pb_last_section.append(pb_lane)
    return pb_last_section


  def construct_signal_overlap(self, pb_lane, pb_signal):
    # lane_overlap_info
    pb_overlap = self.pb_map.overlap.add()
    pb_overlap.id.id = "{}_{}".format(pb_lane.id.id, pb_signal.id.id)
    # lane_overlap_info
    pb_object = pb_overlap.object.add()
    pb_object.id.id = pb_lane.id.id
    # todo(zero): need to complete
    # pb_object.lane_overlap_info.start_s
    # pb_object.lane_overlap_info.end_s

    # signal_overlap_info
    pb_object = pb_overlap.object.add()
    pb_object.id.id = pb_signal.id.id

    # add overlap to lane and signal
    pb_lane_overlap_id = pb_lane.overlap_id.add()
    pb_lane_overlap_id.id = pb_overlap.id.id

    pb_signal_overlap_id = pb_signal.overlap_id.add()
    pb_signal_overlap_id.id = pb_overlap.id.id

  def _construct_signal_stopline(self, last_section_lanes, pb_signal):
    pb_left_lane, pb_right_lane = last_section_lanes[0], last_section_lanes[-1]
    left_segment = pb_left_lane.left_boundary.curve.segment[-1].line_segment
    right_segment = pb_right_lane.right_boundary.curve.segment[-1].line_segment

    sampling_length = global_var.get_element_value("sampling_length")
    index = math.ceil(STOP_LINE_DISTANCE/sampling_length)
    pb_stop_line = pb_signal.stop_line.add()
    pb_segment = pb_stop_line.segment.add()

    if len(left_segment.point) < index:
      point = pb_segment.line_segment.point.add()
      point.CopyFrom(left_segment.point[-1])
      point = pb_segment.line_segment.point.add()
      point.CopyFrom(right_segment.point[-1])
    else:
      point = pb_segment.line_segment.point.add()
      point.CopyFrom(left_segment.point[-index])
      point = pb_segment.line_segment.point.add()
      point.CopyFrom(right_segment.point[-index])


  def convert_signal(self, xodr_road, pb_last_section):
    for signal_reference in xodr_road.signals.signal_references:
      # todo(zero): need implement
      pass

    for signal in xodr_road.signals.signals:
      if signal.is_traffic_light() and pb_last_section:
        # print(signal)
        pb_signal = self.pb_map.signal.add()
        # add road_id to avoid duplication
        pb_signal.id.id = "signal_{}_{}".format(xodr_road.road_id, signal.id)
        # todo(zero): needs to be completed(boundary\subsignal\type)
        self._construct_signal_stopline(pb_last_section, pb_signal)
        for pb_lane in pb_last_section:
          self.construct_signal_overlap(pb_lane, pb_signal)


  def convert_roads(self):
    for _, xodr_road in self.xodr_map.roads.items():
      pb_road = self.pb_map.road.add()
      pb_road.id.id = xodr_road.road_id
      if xodr_road.junction_id != "-1":
        pb_road.junction_id.id = xodr_road.junction_id

      # The definition of road type is inconsistent
      if xodr_road.road_type.road_type is None:
        pb_road.type = map_road_pb2.Road.CITY_ROAD

      xodr_road.generate_reference_line()
      xodr_road.add_offset_to_reference_line()
      xodr_road.add_origin_to_reference_line(self.origin_x, self.origin_y)
      # Todo(zero):
      draw_line(xodr_road.reference_line, 'r', \
        reference_line = True, label = "reference line " + str(pb_road.id.id))

      xodr_road.process_lanes()

      pb_last_section = self.convert_lane(xodr_road, pb_road)
      # Todo(zero): need to complete signal
      self.convert_signal(xodr_road, pb_last_section)

  def _is_valid_junction(self, xodr_junction):
    connecting_roads = set()
    incoming_roads = set()
    for connection in xodr_junction.connections:
      connecting_roads.add(connection.connecting_road)
      incoming_roads.add(connection.incoming_road)

    return len(connecting_roads) != 1 and len(incoming_roads) != 1

  def construct_junction_polygon(self, xodr_junction, type="box"):
    if not self._is_valid_junction(xodr_junction):
      return []

    points = []
    # TODO:: junction boundry 
    if type == "box":
      for road, relation in xodr_junction.connected_roads:
        cross_section = road.get_cross_section(relation)
        if cross_section:
          start, end = cross_section
          points.append([start.x, start.y])
          points.append([end.x, end.y])

          # when point <= 4 convex_hull will not fully covered, so we change to aabb_box
      if len(points) <= 4:
        return aabb_box(points)
      else:
        return convex_hull(points)
    
    elif type == "boundry": 
      left_incomroadlist = []  
      right_incomroadlist = []  
 
      boundary_point_temp=[]
      for junction_connecting_lane in xodr_junction.connections:  
        # connecting_road 
        for lanesection_ in self.xodr_map.roads[junction_connecting_lane.connecting_road].lanes.lane_sections: 
          junction_id = self.xodr_map.roads[junction_connecting_lane.connecting_road].junction_id 
          if junction_id == "-1":
            continue 
          link_name = self.xodr_map.roads[junction_connecting_lane.connecting_road].link
          # print("junction_id: ", junction_id)
          # print("pre: {} sub: {}".format(link_name.predecessor.contact_point, link_name.successor.contact_point))          
          incomRoadID = junction_connecting_lane.connecting_road #incoming_road
          left_boundary_points = []
          right_boundary_points = []  
          
          for lanesectionright in lanesection_.right: # -            
            if  lanesectionright.lane_type != "driving" and lanesectionright.lane_type != "shoulder": 
              # print("laneleft.lane_id.right: ", lanesectionright.lane_id)
              # print("laneleft.type: ", lanesectionright.lane_type)
              if incomRoadID in right_incomroadlist:
                break
              right_incomroadlist.append(incomRoadID) 
              if len(lanesection_.right)>0 and len(lanesection_.left)>0:   
                for points in lanesectionright.right_boundary:              
                      right_boundary_points.append([points.x, points.y])  
              else:
                for points in lanesectionright.right_boundary:              
                      right_boundary_points.append([points.x, points.y])          
                     
                             
          for lanesectionleft in lanesection_.left: # + 
            if lanesectionleft.lane_type != "driving" and lanesectionleft.lane_type != "shoulder":
              # print("laneleft.lane_id.left: ", lanesectionleft.lane_id)
              # print("laneleft.type: ", lanesectionleft.lane_type)  
              if incomRoadID in left_incomroadlist:
                break
              left_incomroadlist.append(incomRoadID)
              if len(lanesection_.right)>0 and len(lanesection_.left)>0:                
                for points in lanesectionleft.left_boundary:                           
                    left_boundary_points.append([points.x, points.y])
              else:
                for points in lanesectionleft.left_boundary:                           
                    left_boundary_points.append([points.x, points.y])
          
          if len(right_boundary_points)>3:
            boundary_point_temp.append([right_boundary_points, link_name, "right"])
          elif len(left_boundary_points)>3:  
            boundary_point_temp.append([left_boundary_points, link_name, "left"])

      if len(boundary_point_temp)==3:
        for i in range(0,len(boundary_point_temp)):
          point_list_temp,link_name_temp, side = boundary_point_temp[i]
          if side == "left":
            if link_name_temp.successor.contact_point == "start" and link_name_temp.predecessor.contact_point == "end":
              boundary_point_temp[i][0]=point_list_temp[::-1]
            else:
              boundary_point_temp[i][0]=point_list_temp
          # else:
          #   if link_name_temp.successor.contact_point == "start" and link_name_temp.predecessor.contact_point == "end":
          #     boundary_point_temp[i][0]=point_list_temp[::-1]
          #   else:
          #     boundary_point_temp[i][0]=point_list_temp

      boundary_points = []  
      point_list_ = []    
      # for point_list,link_name_ in boundary_point_temp:
      if len(boundary_point_temp) == 3:
        point_list_ = boundary_point_temp[0][0]
        for i in range(1,3):
          if boundary_point_temp[0][1].successor.element_id == boundary_point_temp[i][1].predecessor.element_id:
            point_list_.extend(boundary_point_temp[i][0])
            point_list_.extend(boundary_point_temp[3-i][0])
            break
        boundary_points = point_list_
      elif len(boundary_point_temp) == 4:        
        for i in range(0,3):          
          for j in range (i+1, 4):
            if boundary_point_temp[i][1].successor.element_id == boundary_point_temp[j][1].predecessor.element_id:
              temp_point_ = boundary_point_temp[i+1][0]
              temp_name_ = boundary_point_temp[i+1][1]
              boundary_point_temp[i+1][0] = boundary_point_temp[j][0]
              boundary_point_temp[j][0] = temp_point_

              boundary_point_temp[i+1][1] = boundary_point_temp[j][1]
              boundary_point_temp[j][1] = temp_name_
          boundary_points.extend(boundary_point_temp[i][0])
        boundary_points.extend(boundary_point_temp[3][0])
      else:
        for points, _, _ in boundary_point_temp:
          boundary_points.extend(points)
      
      hull = ConvexHull_lib(np.array(boundary_points))


      # return boundary_points
      return concave_hull_lib(boundary_points, length_threshold=0, concavity=5, convex_hull_indexes=hull.vertices)
    else:
      print("---------------------convert method error---------------------")
      return   

  def convert_junctions(self, method="boundry"):
    for _, xodr_junction in self.xodr_map.junctions.items():
      polygon = self.construct_junction_polygon(xodr_junction, method)
      if len(polygon) < 3:
        logging.warning(
          "junction {} polygon size < 3.".format(xodr_junction.junction_id))
        continue

      pb_junction = self.pb_map.junction.add()
      pb_junction.id.id = xodr_junction.junction_id
      for x, y in polygon:
        pb_point = pb_junction.polygon.point.add()
        pb_point.x, pb_point.y, pb_point.z = x, y, 0

  def convert(self, relative=False, junctionbox=False):
    self.convert_header(relative)
    # Don't change the order. "convert_roads" must before "convert_junctions"
    self.convert_roads()
    if junctionbox == False:      
      self.convert_junctions(method="boundry")
    else:      
      self.convert_junctions(method="box")

    # Todo(zero): display xodr map
    if self.output_file_name is None:
      show(need_save=global_var.get_element_value("need_save_figure"), path=self.input_file_name.replace(".xodr", ".png"))


  def save_map(self):
    output_file_name = self.output_file_name
    if output_file_name is None:
      output_file_name = "default"
    write_pb_to_text_file(self.pb_map, output_file_name)
    write_pb_to_bin_file(self.pb_map, output_file_name)



#!/usr/bin/env python

# Copyright 2021 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import imap.editor as editor
import matplotlib.pyplot as plt

from imap.map import Map

fig, ax = plt.subplots(figsize=(20, 20))
# plt.gca().set_aspect('equal', adjustable='box')


def draw(hdmap, lane_id):
    lane_ids = []
    junction_ids = []
    hdmap.draw_lanes(ax, lane_id)
    hdmap.draw_junctions(ax, junction_ids)
    hdmap.draw_signals(ax)
    hdmap.draw_crosswalks(ax)
    hdmap.draw_stop_signs(ax)
    hdmap.draw_yields(ax)

def show_map(map_path, lane_id):
    hdmap=Map()
    hdmap.load(map_path)
    draw(hdmap, lane_id)
    # max windows
    # manager=plt.get_current_fig_manager()
    # manager.window.showMaximized()
    # tight layout
    # todo(zero): why tight layout not work?
    plt.tight_layout()
    plt.axis('equal')
    plt.show()

def add_editor():
  fig.canvas.mpl_connect('button_press_event', editor.on_click)
  fig.canvas.mpl_connect('button_press_event', editor.on_press)
  fig.canvas.mpl_connect('button_release_event', editor.on_release)
  fig.canvas.mpl_connect('pick_event', editor.on_pick)
  fig.canvas.mpl_connect('motion_notify_event', editor.on_motion)


def draw_line(line, color=None, reference_line=False, label=""):
  x = [point.x for point in line]
  y = [point.y for point in line]

  if reference_line:
    ax.plot(x, y, linestyle="dashed", linewidth=10, alpha=0.5, label=label)
  else:
    if color:
      ax.plot(x, y, color, label = label)
    else:
      ax.plot(x, y, label = label)


def show(need_save=False, path=None):
  # show map
  ax.legend()
  ax.axis('equal')
  if need_save:
    plt.savefig(path)
    print(path, "saved")
  plt.show()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="imap_box_up",
    version="0.0.3",
    author="porter",
    author_email="porter.pan@outlook.com",
    description="High-resolution map visualization and conversion tool, opendrive hdmap convert to appollo base map!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/porterpan/imap_box_up",
    project_urls={
        "Bug Tracker": "https://github.com/porterpan/imap_box_up/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    install_requires=[
        'protobuf<=3.19.4',
        'matplotlib',
        'pyproj',
        'record_msg<=0.1.1',
        'concave-hull<=0.0.7',
        'scipy<=1.13.0',
    ],
    entry_points={
        'console_scripts': [
            'imap_box_up = imap.main:main',
        ],
    },
    python_requires=">=3.6",
)
